/*****************************************************************************
 *
 *  Fichier: scoutServoControlLaws.c
 *
 *  Auteur: sf
 *  Date:   Aout 2000
 *  
 *  Description: Les loies de commandes utilisees par  par la tache 
 *               scoutCmdTask du module SCOUT.
 */

#include <stdio.h>
#include <stdlib.h>
#include <portLib.h>
#include <logLib.h>
#include <h2mathLib.h>
#include "server/scoutHeader.h"
#include "scoutConst.h"
#include "scoutServoCntrlLaws.h"

static void errorsGet (double *pdLongit, double *pdTransv, 
		       double *pSdLongit, double *pSdTransv);
static void errorsConfigGet(double *pdLongit, double *pdTransv,
			    double *pdTheta, double *pSdLongit,
			    double *pSdTransv, double *pSdTheta);
static double sinc(double x);

#define scoutCmdTaskPeriod EXEC_TASK_PERIOD(SCOUT_EXECTASK_NUM)

/*========================= SERVO CONTROL LAWS  =========================*/


/******************************************************************************
*
*  scoutPositionControl  -  Asservissement en position du robot
*
*  Description:
*  Cette routine permet d'asservir le robot en position et cap
*
*  Retourne: scoutImportantDriftFlag
*/

BOOL scoutPositionControl (double *vRef, double *wRef)

{
  SCOUT_CMD_PARAM_STR *cmdId = &(scoutDataStrId->cmd);
  SCOUT_REF_STR *refId = &(scoutDataStrId->ref);
  double dLongit;               /* Erreur de deplacement longitudinal */
  double dTransv;               /* Erreur de deplacement transversal */
  double SdLongit;              /* Somme des erreurs longitudinales */
  double SdTransv;              /* Somme des erreurs transverses */
  double sense;                 /* Sens de deplacement */
  double v;                     /* Vitesse lineaire de consigne */
  double w;                     /* Vitesse angulaire de consigne */
  double dv;                    /* Variation de vitesse lineaire */
  double dw;                    /* Variation de vitesse angulaire */
  static double prevV = 0.;      /* Vitesse linaire a la periode precedente */
  static double prevW = 0.;      /* Vitesse angulaire periode precedente */
  int scoutImportantDriftFlag;

  /* Obtenir le sens de deplacement */
  sense = refId->backFlag == SCOUT_BACKWARD_MOTION ? -1 : 1;

  /* Obtenir les ecarts de position et leurs integrales */
  errorsGet (&dLongit, &dTransv, &SdLongit, &SdTransv);

  /* Verifier que le robot suit bien la consigne */
  if (fabs(dLongit) > SCOUT_LONGIT_GAP_MAX ||
      fabs(dTransv) > SCOUT_TRANSV_GAP_MAX) 
    scoutImportantDriftFlag = TRUE;
  else
    scoutImportantDriftFlag = FALSE;
  
  /* Commande de vitesse lineaire a donner au robot */
  v = refId->v + (cmdId->KpLongit*dLongit*refId->linAccelMax)/refId->vmax +
    cmdId->KiLongit * SdLongit;

  /* Limiter l'acceleration lineaire */
  if (fabs (((dv = v - prevV) / scoutCmdTaskPeriod)) > refId->linAccelMax)
    v = prevV +  SIGN (dv) * refId->linAccelMax * scoutCmdTaskPeriod;
  
  /* Limiter la vitesse lineaire */
  if (fabs (v) > refId->vmax)
    v = SIGN (v) * refId->vmax;

  /* Memorisation de la vitesse lineaire */
  prevV = v;
  
  /* Commande de la vitesse angulaire */
  w = refId->w +
    sense *(cmdId->KpTransv*dTransv*refId->angAccelMax/refId->wmax +
	    cmdId->KiTransv*SdTransv);
  
  /* Limiter l'acceleration angulaire */
  if (fabs ((dw = w - prevW) / scoutCmdTaskPeriod) > refId->angAccelMax)
    w = prevW + SIGN (dw) * refId->angAccelMax * scoutCmdTaskPeriod;

  /* Limiter la vitesse angulaire */
  if (fabs (w) >  refId->wmax)
    w = SIGN (w) * refId->wmax;

  /* Memorisation de la vitesse angulaire */
  prevW = w;

  /* Imposer au robot les vitesses lineaire et angulaire de consigne */
  *vRef = v;
  *wRef = w;
  return scoutImportantDriftFlag;
}


/******************************************************************************
*
*  scoutConfigurationControl  -  Asservissement en configuration du robot
*
*  Description:
*  Cette routine permet d'asservir le robot en position et cap
*
*  Retourne: scoutImportantDriftFlag
*/

STATUS scoutConfigurationControl (double *vRef, double *wRef)

{
  SCOUT_CMD_PARAM_STR *cmdId = &(scoutDataStrId->cmd);
  SCOUT_REF_STR *refId = &(scoutDataStrId->ref);
  double dLongit;               /* Erreur de deplacement longitudinal */
  double dTransv;               /* Erreur de deplacement transversal */
  double dTheta;                /* Erreur angulaire */
  double SdLongit;              /* Somme des erreurs longitudinales */
  double SdTransv;              /* Somme des erreurs transverses */
  double SdTheta;               /* Somme des erreurs angulaires */
  double v, v0 = refId->v;      /* Vitesse lineaire de consigne */
  double w;                     /* Vitesse angulaire de consigne */
  double dv;                    /* Variation de vitesse lineaire */
  double dw;                    /* Variation de vitesse angulaire */
  static double prevV = 0.;      /* Vitesse linaire a la periode precedente */
  static double prevW = 0.;      /* Vitesse angulaire periode precedente */

  int scoutImportantDriftFlag;

  /* Obtenir les ecarts de position et leurs integrales */
  errorsConfigGet(&dLongit, &dTransv, &dTheta, 
		  &SdLongit, &SdTransv, &SdTheta);

  /* Verifier que le robot suit bien la consigne */
  if (fabs(dLongit) > SCOUT_LONGIT_GAP_MAX ||
      fabs(dTransv) > SCOUT_TRANSV_GAP_MAX ||
      fabs(dTheta)  > SCOUT_ANG_GAP_MAX) 
    scoutImportantDriftFlag = TRUE;
  else
    scoutImportantDriftFlag = FALSE;
					  
  /* Commande de vitesse lineaire a donner au robot */
  v = v0*(cos(dTheta)+cmdId->KpLongit*dLongit
	  +cmdId->KiLongit*SdLongit);
  /* Limiter l'acceleration lineaire */
  if (fabs (((dv = v - prevV) / scoutCmdTaskPeriod)) > refId->linAccelMax)
    v = prevV +  SIGN (dv) * refId->linAccelMax * scoutCmdTaskPeriod;
  
  /* Limiter la vitesse lineaire */
  if (fabs (v) > refId->vmax)
    v = SIGN (v) * refId->vmax;

  /* Memorisation de la vitesse lineaire */
  prevV = v;
  
  /* Commande de la vitesse angulaire */
  w = refId->w+v0*(cmdId->KpAng*dTheta+cmdId->KiAng*SdTheta+
		   sinc(dTheta)*(cmdId->KpTransv*dTransv+
				 cmdId->KiTransv*SdTransv));

  /* Limiter l'acceleration angulaire */
  if (fabs ((dw = w - prevW) / scoutCmdTaskPeriod) > refId->angAccelMax)
    w = prevW + SIGN (dw) * refId->angAccelMax * scoutCmdTaskPeriod;

  /* Limiter la vitesse angulaire */
  if (fabs (w) >  refId->wmax)
    w = SIGN (w) * refId->wmax;

  /* Memorisation de la vitesse angulaire */
  prevW = w;

  /* Imposer au robot les vitesses lineaire et angulaire de consigne */
  *vRef = v;
  *wRef = w;
  return scoutImportantDriftFlag;
}


/******************************************************************************
*
*  scoutSpeedControl  -  Asservissement en position du robot
*
*  Description:
*  Cette routine permet d'asservir le robot en position et cap
*
*  Retourne: scoutImportantDriftFlag
*/

BOOL scoutSpeedControl (SCOUT_SPEED_REF_STR *refId, 
			SCOUT_SPEED_REF_STR *refOutId)
{
#if 0
  double v;                     /* Vitesse lineaire de consigne */
  double w;                     /* Vitesse angulaire de consigne */
  double dv;                    /* Variation de vitesse lineaire */
  double dw;                    /* Variation de vitesse angulaire */
#endif
  static double prevV = 0.;      /* Vitesse linaire a la periode precedente */
  static double prevW = 0.;      /* Vitesse angulaire periode precedente */
  int scoutImportantDriftFlag=FALSE;

  *refOutId = *refId;

  /* XXXX va disparaitre car redondant avec scoutVelocityControl */
  /* NON, pas forcement redondant car peut permettre de limiter 
     les vitesses en restant sur la trajectoire (ie, en conservant 
     le rapport v/w). Regarder ce qui a ete fait dans lscout de lama. */
#if 0
  /* Commande de vitesse lineaire a donner au robot */
  v = refId->v;

  /* Limiter l'acceleration lineaire */
  if (fabs (((dv = v - prevV) / scoutCmdTaskPeriod)) > AMAX_LIN_ROBOT)
    v = prevV +  SIGN (dv) * AMAX_LIN_ROBOT * scoutCmdTaskPeriod;
  
  /* Limiter la vitesse lineaire */
  if (fabs (v) > VMAX_ROBOT)
    v = SIGN (v) * VMAX_ROBOT;

  /* Memorisation de la vitesse lineaire */
  prevV = v;
  
  /* Commande de la vitesse angulaire */
  w = refId->w;

  /* Limiter l'acceleration angulaire */
  if (fabs ((dw = w - prevW) / scoutCmdTaskPeriod) > AMAX_ANG_ROBOT)
    w = prevW + SIGN (dw) * AMAX_ANG_ROBOT * scoutCmdTaskPeriod;

  /* Limiter la vitesse angulaire */
  if (fabs (w) >  WMAX_ROBOT)
    w = SIGN (w) * WMAX_ROBOT;

  /* Memorisation de la vitesse angulaire */
  prevW = w;

  refOutId->v = v;
  refOutId->w = w;
#endif

  prevV = refId->v;
  prevW = refId->w;

  return scoutImportantDriftFlag;
}




/*============================= LOCAL FUNCTIONS  =========================*/


/*****************************************************************************
*
*  errorsGet  -  Obtenir les ecarts de position du robot
*
*  Description:
*  Cette fonction permet d'obtenir les ecarts longitudinal et transversal
*  du point commande' par rapport a la position de consigne. Cette fonction
*  retourne aussi les integrales des erreurs longitudinale et transversale.
*
*  Retourne: Neant
*/

static void errorsGet (double *pdLongit, /* Ou` mettre l'ecart longitudinal */
		      double *pdTransv,  /* Ou` mettre l'ecart transversal */
		      double *pSdLongit, /* Ou` mettre somme ecarts longit. */
		      double *pSdTransv) /* Ou` mettre somme ecarts transv. */

{
  SCOUT_CMD_PARAM_STR *cmdId = &(scoutDataStrId->cmd);
  SCOUT_REF_STR *refId = &(scoutDataStrId->ref);
  double dxA, dyA;              /* Erreurs position par rapport rep atelier */
  double dLongit;               /* Erreur de deplacement longitudinal */
  double dTransv;               /* Erreur de deplacement transversal */
  static double SdLongit = 0.;   /* Somme des erreurs longitudinales */
  static double SdTransv = 0.;   /* Somme des erreurs transverses */
  double c, s;
  static SCOUT_POS_STR *robotId;


  /* Donne'es positionnelles sur les quelles on s'asservit */
  robotId = &(scoutDataStrId->odo.robot);
  c = cos(robotId->theta);
  s = sin(robotId->theta);

  /* Obtenir les erreurs de position par rapport au rep atelier */
  dxA = refId->x - robotId->x;
  dyA = refId->y - robotId->y;
  
  /* Obtenir les erreurs de deplacement transversal et longitudinal */
  *pdLongit = dLongit =   dxA * c + dyA * s;
  *pdTransv = dTransv = - dxA * s + dyA * c;

  /* Actualiser l'integrale des erreurs longitudinales */
  if (fabs (dLongit) < cmdId->longitErrorMax) {

    /* Actualiser la somme des erreurs */
    SdLongit += dLongit*scoutCmdTaskPeriod;
    
    /* La limiter */
    if (fabs (SdLongit) > cmdId->longitMaxErrSum)
      *pSdLongit = SdLongit = SIGN(SdLongit) * cmdId->longitMaxErrSum;
    else 
      *pSdLongit = SdLongit;
  }
  else 
    *pSdLongit = SdLongit = 0.;

  /* Actualiser l'integrale des erreurs transversales */
  if (fabs (dTransv) < cmdId->transvErrorMax) {
    
    /* Actualiser la somme des erreurs */
    SdTransv += dTransv*scoutCmdTaskPeriod;
    
    /* La limiter */
    if (fabs (SdTransv) > cmdId->transvMaxErrSum)
      *pSdTransv = SdTransv = SIGN(SdTransv) * cmdId->transvMaxErrSum;
    else
      *pSdTransv = SdTransv;
  }
  else 
    *pSdTransv = SdTransv = 0.;
}


/*****************************************************************************
*
*  errorsConfigGet  -  Obtenir les ecarts de position du robot
*
*  Description:
*  Cette fonction permet d'obtenir les ecarts longitudinal et transversal
*  du centre du robot par rapport a la position de consigne. Cette fonction
*  retourne aussi les integrales des erreurs longitudinale et transversale.
*
*  Retourne: Neant
*/

static void errorsConfigGet(double *pdLongit, /* Ecart longitudinal */
		       double *pdTransv,  /* Ecart transversal */
		       double *pdTheta,   /* Ecart angulaire */
		       double *pSdLongit, /* Somme ecarts longit. */
		       double *pSdTransv, /* Somme ecarts transv. */
		       double *pSdTheta) /* Somme ecarts ang. */

{
  SCOUT_CMD_PARAM_STR *cmdId = &(scoutDataStrId->cmd);
  SCOUT_REF_STR *refId = &(scoutDataStrId->ref);
  double dxA, dyA;              /* Erreurs position par rapport rep atelier */
  double dLongit;               /* Erreur de deplacement longitudinal */
  double dTransv;               /* Erreur de deplacement transversal */
  double dTheta=0;                /* Erreur de deplacement angulaire */
  static double SdLongit = 0.;   /* Somme des erreurs longitudinales */
  static double SdTransv = 0.;   /* Somme des erreurs transverses */
  static double SdTheta = 0.;      /* Somme des erreurs angulaires */
  double c, s;
  static SCOUT_POS_STR *robotId;


  /* Donne'es positionnelles sur lesquelles on s'asservi */
  robotId = &(scoutDataStrId->odo.robot);
  c = cos(robotId->theta);
  s = sin(robotId->theta);

  /* Obtenir les erreurs de position par rapport au rep atelier */
  /* point de controle */
  dxA = refId->x - robotId->xC;
  dyA = refId->y - robotId->yC;
  
  /* angles */
  *pdTheta = angleLimit(refId->theta - robotId->theta);

  /* Obtenir les erreurs de deplacement transversal et longitudinal */
  *pdLongit = dLongit =   dxA * c + dyA * s;
  *pdTransv = dTransv = - dxA * s + dyA * c;

  /* Actualiser l'integrale des erreurs longitudinales */
  if (fabs (dLongit) < cmdId->longitErrorMax) {

    /* Actualiser la somme des erreurs */
    SdLongit += dLongit*scoutCmdTaskPeriod;
    
    /* La limiter */
    if (fabs (SdLongit) > cmdId->longitMaxErrSum)
      *pSdLongit = SdLongit = cmdId->longitMaxErrSum;
    else 
      *pSdLongit = SdLongit;
  }
  else 
    *pSdLongit = SdLongit = 0.;

  /* Actualiser l'integrale des erreurs transversales */
  if (fabs (dTransv) < cmdId->transvErrorMax) {
    
    /* Actualiser la somme des erreurs */
    SdTransv += dTransv*scoutCmdTaskPeriod;
    
    /* La limiter */
    if (fabs (SdTransv) > cmdId->transvMaxErrSum)
      *pSdTransv = SdTransv = cmdId->transvMaxErrSum;
    else
      *pSdTransv = SdTransv;
  }
  else 
    *pSdTransv = SdTransv = 0.;

  /* Actualiser l'integrale des erreurs angulaires */
  if (fabs (dTheta) < cmdId->angErrorMax) {
    
    /* Actualiser la somme des erreurs */
    SdTheta += dTheta*scoutCmdTaskPeriod;
    
    /* La limiter */
    if (fabs (SdTheta) > cmdId->angMaxErrSum)
      *pSdTheta = SdTheta = cmdId->angMaxErrSum;
    else
      *pSdTheta = SdTheta;
  }
  else 
    *pSdTheta = SdTheta = 0.;
}

/**
 **  sinc  
 **
 **/
static double sinc(double x)
{
  double y=1.0;

  if (fabs(x)>1e-8)
    {
      y = sin(x)/x;
    }
  return y;
}
